from __future__ import annotations

from fnmatch import fnmatch
from pathlib import Path

from .diffing import compare_schema_snapshots
from .infer import infer_schema
from .models import (
    BaselineEntry,
    BaselineManifest,
    CompareOptions,
    ComparisonReport,
    DirectoryEntry,
    DirectoryReport,
    SchemaSnapshot,
    Severity,
)

SUPPORTED_SCAN_SUFFIXES = {
    ".avsc",
    ".csv",
    ".json",
    ".jsonl",
    ".parquet",
    ".proto",
    ".yaml",
    ".yml",
}


def inspect_file(path: str | Path, sample_size: int = 10_000) -> SchemaSnapshot:
    return infer_schema(Path(path), sample_size=sample_size)


def snapshot_file(path: str | Path, sample_size: int = 10_000) -> SchemaSnapshot:
    return inspect_file(path, sample_size=sample_size)


def compare_files(
    old_path: str | Path, new_path: str | Path, options: CompareOptions | None = None
) -> ComparisonReport:
    compare_options = options or CompareOptions()
    old_snapshot = infer_schema(Path(old_path), sample_size=compare_options.sample_size)
    new_snapshot = infer_schema(Path(new_path), sample_size=compare_options.sample_size)
    return compare_schema_snapshots(old_snapshot, new_snapshot, compare_options)


def compare_snapshots(
    old_snapshot: SchemaSnapshot,
    new_snapshot: SchemaSnapshot,
    options: CompareOptions | None = None,
) -> ComparisonReport:
    return compare_schema_snapshots(old_snapshot, new_snapshot, options or CompareOptions())


def load_snapshot(path: str | Path) -> SchemaSnapshot:
    snapshot_path = Path(path)
    return SchemaSnapshot.model_validate_json(snapshot_path.read_text(encoding="utf-8"))


def save_snapshot(snapshot: SchemaSnapshot, path: str | Path) -> None:
    Path(path).write_text(snapshot.model_dump_json(indent=2), encoding="utf-8")


def scan_directories(
    old_root: str | Path,
    new_root: str | Path,
    options: CompareOptions | None = None,
    pattern: str = "*",
) -> DirectoryReport:
    compare_options = options or CompareOptions()
    old_dir = Path(old_root)
    new_dir = Path(new_root)
    old_files = _collect_supported_files(old_dir, pattern=pattern)
    new_files = _collect_supported_files(new_dir, pattern=pattern)
    entries: list[DirectoryEntry] = []

    for relative_path in sorted(old_files.keys() | new_files.keys()):
        old_path = old_files.get(relative_path)
        new_path = new_files.get(relative_path)
        if old_path is None and new_path is not None:
            entries.append(
                DirectoryEntry(
                    relative_path=relative_path,
                    severity=Severity.WARNING,
                    status="added_file",
                    summary=f"new file added: {relative_path}",
                )
            )
            continue
        if old_path is not None and new_path is None:
            entries.append(
                DirectoryEntry(
                    relative_path=relative_path,
                    severity=Severity.BREAKING,
                    status="removed_file",
                    summary=f"file removed: {relative_path}",
                )
            )
            continue
        assert old_path is not None and new_path is not None
        report = compare_files(old_path, new_path, options=compare_options)
        entries.append(
            DirectoryEntry(
                relative_path=relative_path,
                severity=report.overall,
                status="compared",
                summary=f"{relative_path}: {report.overall.value}",
                report=report,
            )
        )

    return DirectoryReport(old_root=str(old_dir), new_root=str(new_dir), entries=entries)


def capture_baseline(
    source_root: str | Path,
    baseline_root: str | Path,
    sample_size: int = 10_000,
    pattern: str = "*",
) -> BaselineManifest:
    source_dir = Path(source_root)
    baseline_dir = Path(baseline_root)
    snapshots_dir = baseline_dir / "snapshots"
    snapshots_dir.mkdir(parents=True, exist_ok=True)

    entries: list[BaselineEntry] = []
    for relative_path, file_path in sorted(
        _collect_supported_files(source_dir, pattern=pattern).items()
    ):
        snapshot = infer_schema(file_path, sample_size=sample_size)
        snapshot_path = snapshots_dir / f"{relative_path}.schema.json"
        snapshot_path.parent.mkdir(parents=True, exist_ok=True)
        save_snapshot(snapshot, snapshot_path)
        entries.append(
            BaselineEntry(
                relative_path=relative_path,
                snapshot_path=str(snapshot_path.relative_to(baseline_dir).as_posix()),
                source_format=snapshot.source_format,
            )
        )

    manifest = BaselineManifest(
        source_root=str(source_dir), sample_size=sample_size, entries=entries
    )
    (baseline_dir / "schemaglow-baseline.json").write_text(
        manifest.model_dump_json(indent=2), encoding="utf-8"
    )
    return manifest


def load_baseline_manifest(path: str | Path) -> BaselineManifest:
    manifest_root = Path(path)
    manifest_path = (
        manifest_root / "schemaglow-baseline.json" if manifest_root.is_dir() else manifest_root
    )
    return BaselineManifest.model_validate_json(manifest_path.read_text(encoding="utf-8"))


def check_baseline(
    baseline_root: str | Path,
    candidate_root: str | Path,
    options: CompareOptions | None = None,
    pattern: str = "*",
) -> DirectoryReport:
    compare_options = options or CompareOptions()
    baseline_dir = Path(baseline_root)
    candidate_dir = Path(candidate_root)
    manifest = load_baseline_manifest(baseline_dir)
    candidate_files = _collect_supported_files(candidate_dir, pattern=pattern)
    manifest_entries = {
        entry.relative_path: entry
        for entry in manifest.entries
        if _matches_pattern(entry.relative_path, pattern)
    }
    entries: list[DirectoryEntry] = []

    for relative_path in sorted(manifest_entries):
        entry = manifest_entries[relative_path]
        candidate_path = candidate_files.pop(relative_path, None)
        snapshot_path = baseline_dir / entry.snapshot_path
        if candidate_path is None:
            entries.append(
                DirectoryEntry(
                    relative_path=relative_path,
                    severity=Severity.BREAKING,
                    status="missing_candidate",
                    summary=f"baseline file missing from candidate tree: {relative_path}",
                )
            )
            continue
        baseline_snapshot = load_snapshot(snapshot_path)
        candidate_snapshot = infer_schema(candidate_path, sample_size=compare_options.sample_size)
        report = compare_snapshots(baseline_snapshot, candidate_snapshot, options=compare_options)
        entries.append(
            DirectoryEntry(
                relative_path=relative_path,
                severity=report.overall,
                status="baseline_check",
                summary=f"{relative_path}: {report.overall.value}",
                report=report,
            )
        )

    for relative_path in sorted(candidate_files):
        entries.append(
            DirectoryEntry(
                relative_path=relative_path,
                severity=Severity.WARNING,
                status="unexpected_candidate",
                summary=f"candidate file not present in baseline: {relative_path}",
            )
        )

    return DirectoryReport(
        old_root=str(baseline_dir),
        new_root=str(candidate_dir),
        entries=entries,
    )


def _collect_supported_files(root: Path, pattern: str) -> dict[str, Path]:
    files: dict[str, Path] = {}
    for path in root.rglob("*"):
        if not path.is_file() or path.suffix.lower() not in SUPPORTED_SCAN_SUFFIXES:
            continue
        relative_path = path.relative_to(root).as_posix()
        if not _matches_pattern(relative_path, pattern):
            continue
        files[relative_path] = path
    return files


def _matches_pattern(relative_path: str, pattern: str) -> bool:
    return (
        pattern == "*" or fnmatch(relative_path, pattern) or fnmatch(relative_path, f"**/{pattern}")
    )


__author__ = "gadwant"



"""SchemaGlow package."""

from .service import (
    capture_baseline,
    check_baseline,
    compare_files,
    compare_snapshots,
    inspect_file,
    scan_directories,
    snapshot_file,
)

__all__ = [
    "capture_baseline",
    "check_baseline",
    "compare_files",
    "compare_snapshots",
    "inspect_file",
    "scan_directories",
    "snapshot_file",
]

__author__ = "gadwant"
__version__ = "1.0.0"



from __future__ import annotations

import json
from pathlib import Path

from schemaglow.models import CompareOptions, Severity
from schemaglow.renderers import (
    render_baseline_manifest_text,
    render_directory_html,
    render_directory_json,
    render_directory_markdown,
    render_directory_text,
    render_html_report,
    render_json_report,
    render_markdown_report,
    render_snapshot_text,
    render_text_report,
    write_report,
)
from schemaglow.service import (
    capture_baseline,
    check_baseline,
    compare_files,
    compare_snapshots,
    load_snapshot,
    save_snapshot,
    scan_directories,
    snapshot_file,
)


def _write(path: Path, content: str) -> Path:
    path.write_text(content, encoding="utf-8")
    return path


def test_renderers_emit_all_supported_report_formats(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "id,total\n1,10\n2,20\n")
    new_file = _write(tmp_path / "new.csv", "id,total,status\n1,10,open\n2,20,\n")

    report = compare_files(old_file, new_file, options=CompareOptions())

    text_output = render_text_report(report)
    json_output = render_json_report(report)
    markdown_output = render_markdown_report(report)
    html_output = render_html_report(report)

    assert "SchemaGlow Report" in text_output
    assert "added optional field: status (string)" in text_output
    assert json.loads(json_output)["overall"] == Severity.SAFE.value
    assert "# SchemaGlow Report" in markdown_output
    assert "added optional field: status (string)" in markdown_output
    assert "<html" in html_output.lower()
    assert "campaign_id" not in html_output


def test_snapshot_save_load_and_compare_roundtrip(tmp_path: Path) -> None:
    source_file = _write(tmp_path / "source.json", '{"order":{"id":1,"total":12.5}}')
    snapshot_path = tmp_path / "snapshot.json"

    snapshot = snapshot_file(source_file)
    save_snapshot(snapshot, snapshot_path)
    loaded_snapshot = load_snapshot(snapshot_path)
    comparison = compare_snapshots(snapshot, loaded_snapshot)

    assert loaded_snapshot.source_format == "json"
    assert loaded_snapshot.field_map["order.total"].type_name == "number"
    assert comparison.overall == Severity.SAFE
    assert comparison.events[0].code == "no_change"


def test_render_snapshot_text_and_write_report(tmp_path: Path) -> None:
    source_file = _write(tmp_path / "records.jsonl", '{"user":{"id":1}}\n{"user":{"id":2}}\n')
    report_path = tmp_path / "inspect.txt"

    snapshot = snapshot_file(source_file)
    inspect_output = render_snapshot_text(snapshot)
    write_report(report_path, inspect_output)

    assert "SchemaGlow Inspect" in inspect_output
    assert "- user.id: integer (required)" in inspect_output
    assert report_path.read_text(encoding="utf-8") == inspect_output


def test_renderers_include_ignored_fields_and_before_after_details(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "id,_loaded_at\n1,2026-03-09\n")
    new_file = _write(tmp_path / "new.csv", "id,_loaded_at\none,2026-03-10\n")

    report = compare_files(
        old_file, new_file, options=CompareOptions(ignore_fields=r"^_loaded_at$")
    )

    text_output = render_text_report(report)
    markdown_output = render_markdown_report(report)

    assert "ignored: _loaded_at" in text_output
    assert "before: integer / required" in text_output
    assert "## Ignored Fields" in markdown_output
    assert "- Before: `integer` / required" in markdown_output


def test_scan_directories_and_baseline_check(tmp_path: Path) -> None:
    old_dir = tmp_path / "old"
    new_dir = tmp_path / "new"
    baseline_dir = tmp_path / "baseline"
    old_dir.mkdir()
    new_dir.mkdir()

    _write(old_dir / "orders.csv", "id,total\n1,10\n")
    _write(new_dir / "orders.csv", "id,total,status\n1,10,\n")
    _write(new_dir / "extra.json", '{"id": 1}')

    scan_report = scan_directories(old_dir, new_dir, options=CompareOptions())
    manifest = capture_baseline(old_dir, baseline_dir)
    baseline_report = check_baseline(baseline_dir, new_dir, options=CompareOptions())

    assert manifest.entries[0].relative_path == "orders.csv"
    assert scan_report.overall == Severity.WARNING
    assert any(entry.status == "added_file" for entry in scan_report.entries)
    assert baseline_report.overall == Severity.WARNING
    assert any(entry.status == "unexpected_candidate" for entry in baseline_report.entries)


def test_directory_renderers_and_baseline_manifest_text(tmp_path: Path) -> None:
    old_dir = tmp_path / "old"
    new_dir = tmp_path / "new"
    baseline_dir = tmp_path / "baseline"
    old_dir.mkdir()
    new_dir.mkdir()

    _write(old_dir / "orders.csv", "id,total\n1,10\n")
    _write(new_dir / "orders.csv", "id,total\n1,10.5\n")

    scan_report = scan_directories(old_dir, new_dir, options=CompareOptions(strict=True))
    manifest = capture_baseline(old_dir, baseline_dir)

    text_output = render_directory_text(scan_report)
    json_output = render_directory_json(scan_report)
    markdown_output = render_directory_markdown(scan_report)
    html_output = render_directory_html(scan_report)
    manifest_output = render_baseline_manifest_text(manifest, baseline_dir)

    assert "SchemaGlow Scan Report" in text_output
    assert '"overall": "WARNING"' in json_output
    assert "# SchemaGlow Scan Report" in markdown_output
    assert "<html" in html_output.lower()
    assert "SchemaGlow Baseline" in manifest_output
    assert "orders.csv" in manifest_output



from __future__ import annotations

from pathlib import Path

import typer
from rich.console import Console

from .models import CompareOptions
from .renderers import (
    render_baseline_manifest_text,
    render_directory_html,
    render_directory_json,
    render_directory_markdown,
    render_directory_text,
    render_html_report,
    render_json_report,
    render_markdown_report,
    render_snapshot_text,
    render_text_report,
    write_report,
)
from .service import (
    capture_baseline,
    check_baseline,
    compare_files,
    compare_snapshots,
    inspect_file,
    load_baseline_manifest,
    load_snapshot,
    save_snapshot,
    scan_directories,
    snapshot_file,
)

app = typer.Typer(help="Human-friendly schema diff for CSV, JSON, JSONL, and Parquet.")
baseline_app = typer.Typer(help="Capture and check baseline contract snapshots.")
app.add_typer(baseline_app, name="baseline")
console = Console(soft_wrap=True, force_terminal=False, color_system=None)


def _emit_report(report_content: str, report_path: Path | None) -> None:
    if report_path is None:
        console.print(report_content)
        return
    write_report(report_path, report_content)
    console.print(f"report written to {report_path}")


def _emit_comparison_output(format: str, comparison: object) -> None:
    if format == "json":
        if hasattr(comparison, "entries"):
            console.print(render_directory_json(comparison))  # type: ignore[arg-type]
        else:
            console.print(render_json_report(comparison))  # type: ignore[arg-type]
        return
    if hasattr(comparison, "entries"):
        console.print(render_directory_text(comparison))  # type: ignore[arg-type]
        return
    console.print(render_text_report(comparison))  # type: ignore[arg-type]


def _emit_comparison_report(
    report: str | None, report_path: Path | None, comparison: object
) -> None:
    if report == "markdown":
        content = (
            render_directory_markdown(comparison)  # type: ignore[arg-type]
            if hasattr(comparison, "entries")
            else render_markdown_report(comparison)  # type: ignore[arg-type]
        )
        _emit_report(content, report_path)
    elif report == "html":
        content = (
            render_directory_html(comparison)  # type: ignore[arg-type]
            if hasattr(comparison, "entries")
            else render_html_report(comparison)  # type: ignore[arg-type]
        )
        _emit_report(content, report_path)
    elif report is not None:
        raise typer.BadParameter("report must be one of: markdown, html")


@app.command()
def diff(
    old_file: Path = typer.Argument(..., help="Baseline data file path."),
    new_file: Path = typer.Argument(..., help="Candidate data file path."),
    format: str = typer.Option("text", "--format", help="Output format: text or json."),
    report: str | None = typer.Option(
        None, "--report", help="Optional report format: markdown or html."
    ),
    report_path: Path | None = typer.Option(
        None, "--report-path", help="Write the generated report to a file."
    ),
    sample: int = typer.Option(10_000, "--sample", help="Maximum rows/items to inspect."),
    ignore_order: bool = typer.Option(
        False, "--ignore-order", help="Suppress column order change events."
    ),
    ignore_fields: str | None = typer.Option(
        None, "--ignore-fields", help="Regex for field paths to ignore."
    ),
    strict: bool = typer.Option(
        False, "--strict", help="Treat numeric widening as warning instead of safe."
    ),
    rename_heuristics: bool = typer.Option(
        False, "--rename-heuristics", help="Detect likely field renames."
    ),
) -> None:
    options = CompareOptions(
        sample_size=sample,
        ignore_order=ignore_order,
        ignore_fields=ignore_fields,
        strict=strict,
        rename_heuristics=rename_heuristics,
    )
    comparison = compare_files(old_file, new_file, options=options)
    _emit_comparison_output(format, comparison)
    _emit_comparison_report(report, report_path, comparison)


@app.command()
def inspect(
    file: Path = typer.Argument(..., help="Data file to inspect."),
    format: str = typer.Option("text", "--format", help="Output format: text or json."),
    sample: int = typer.Option(10_000, "--sample", help="Maximum rows/items to inspect."),
) -> None:
    snapshot = inspect_file(file, sample_size=sample)
    if format == "json":
        console.print(snapshot.model_dump_json(indent=2))
    else:
        console.print(render_snapshot_text(snapshot))


@app.command()
def snapshot(
    file: Path = typer.Argument(..., help="Data file to snapshot."),
    output: Path = typer.Option(..., "-o", "--output", help="Snapshot JSON path."),
    sample: int = typer.Option(10_000, "--sample", help="Maximum rows/items to inspect."),
) -> None:
    schema_snapshot = snapshot_file(file, sample_size=sample)
    save_snapshot(schema_snapshot, output)
    console.print(f"snapshot written to {output}")


@app.command()
def compare(
    old_schema: Path = typer.Argument(..., help="Baseline schema snapshot JSON."),
    new_schema: Path = typer.Argument(..., help="Candidate schema snapshot JSON."),
    format: str = typer.Option("text", "--format", help="Output format: text or json."),
    ignore_order: bool = typer.Option(
        False, "--ignore-order", help="Suppress column order change events."
    ),
    ignore_fields: str | None = typer.Option(
        None, "--ignore-fields", help="Regex for field paths to ignore."
    ),
    strict: bool = typer.Option(
        False, "--strict", help="Treat numeric widening as warning instead of safe."
    ),
    rename_heuristics: bool = typer.Option(
        False, "--rename-heuristics", help="Detect likely field renames."
    ),
) -> None:
    options = CompareOptions(
        ignore_order=ignore_order,
        ignore_fields=ignore_fields,
        strict=strict,
        rename_heuristics=rename_heuristics,
    )
    comparison = compare_snapshots(
        load_snapshot(old_schema), load_snapshot(new_schema), options=options
    )
    _emit_comparison_output(format, comparison)


@app.command()
def scan(
    old_root: Path = typer.Argument(..., help="Baseline directory path."),
    new_root: Path = typer.Argument(..., help="Candidate directory path."),
    format: str = typer.Option("text", "--format", help="Output format: text or json."),
    report: str | None = typer.Option(
        None, "--report", help="Optional report format: markdown or html."
    ),
    report_path: Path | None = typer.Option(
        None, "--report-path", help="Write the generated report to a file."
    ),
    pattern: str = typer.Option(
        "*", "--pattern", help="Glob pattern for files to include in the scan."
    ),
    sample: int = typer.Option(10_000, "--sample", help="Maximum rows/items to inspect."),
    ignore_order: bool = typer.Option(
        False, "--ignore-order", help="Suppress column order change events."
    ),
    ignore_fields: str | None = typer.Option(
        None, "--ignore-fields", help="Regex for field paths to ignore."
    ),
    strict: bool = typer.Option(
        False, "--strict", help="Treat numeric widening as warning instead of safe."
    ),
    rename_heuristics: bool = typer.Option(
        False, "--rename-heuristics", help="Detect likely field renames."
    ),
) -> None:
    options = CompareOptions(
        sample_size=sample,
        ignore_order=ignore_order,
        ignore_fields=ignore_fields,
        strict=strict,
        rename_heuristics=rename_heuristics,
    )
    report_data = scan_directories(old_root, new_root, options=options, pattern=pattern)
    _emit_comparison_output(format, report_data)
    _emit_comparison_report(report, report_path, report_data)


@baseline_app.command("capture")
def baseline_capture(
    source_root: Path = typer.Argument(..., help="Directory to snapshot into a baseline."),
    output: Path = typer.Option(..., "-o", "--output", help="Baseline directory."),
    pattern: str = typer.Option(
        "*", "--pattern", help="Glob pattern for files to include in the baseline."
    ),
    sample: int = typer.Option(10_000, "--sample", help="Maximum rows/items to inspect."),
) -> None:
    manifest = capture_baseline(source_root, output, sample_size=sample, pattern=pattern)
    console.print(render_baseline_manifest_text(manifest, output))


@baseline_app.command("check")
def baseline_check(
    baseline_root: Path = typer.Argument(
        ..., help="Baseline directory containing manifest and snapshots."
    ),
    candidate_root: Path = typer.Argument(..., help="Candidate directory to validate."),
    format: str = typer.Option("text", "--format", help="Output format: text or json."),
    report: str | None = typer.Option(
        None, "--report", help="Optional report format: markdown or html."
    ),
    report_path: Path | None = typer.Option(
        None, "--report-path", help="Write the generated report to a file."
    ),
    pattern: str = typer.Option(
        "*", "--pattern", help="Glob pattern for files to include in the check."
    ),
    sample: int = typer.Option(10_000, "--sample", help="Maximum rows/items to inspect."),
    ignore_order: bool = typer.Option(
        False, "--ignore-order", help="Suppress column order change events."
    ),
    ignore_fields: str | None = typer.Option(
        None, "--ignore-fields", help="Regex for field paths to ignore."
    ),
    strict: bool = typer.Option(
        False, "--strict", help="Treat numeric widening as warning instead of safe."
    ),
    rename_heuristics: bool = typer.Option(
        False, "--rename-heuristics", help="Detect likely field renames."
    ),
) -> None:
    load_baseline_manifest(baseline_root)
    options = CompareOptions(
        sample_size=sample,
        ignore_order=ignore_order,
        ignore_fields=ignore_fields,
        strict=strict,
        rename_heuristics=rename_heuristics,
    )
    report_data = check_baseline(
        baseline_root,
        candidate_root,
        options=options,
        pattern=pattern,
    )
    _emit_comparison_output(format, report_data)
    _emit_comparison_report(report, report_path, report_data)


def main() -> None:  # pragma: no cover
    app()


if __name__ == "__main__":  # pragma: no cover
    main()

__author__ = "gadwant"



from __future__ import annotations

import shutil
from pathlib import Path

from schemaglow.models import CompareOptions, Severity
from schemaglow.service import (
    capture_baseline,
    check_baseline,
    compare_files,
    inspect_file,
    scan_directories,
)

FIXTURES = Path(__file__).resolve().parents[1] / "fixtures"
DOWNLOADED = FIXTURES / "downloaded"
MANUAL = FIXTURES / "manual"


def test_downloaded_samples_cover_every_supported_input_type() -> None:
    expected = {
        "seattle-weather.csv": ("csv", "date"),
        "miserables.json": ("json", "nodes[].name"),
        "example.jsonl": ("jsonl", "result.sequenceNumber"),
        "alltypes_plain.parquet": ("parquet", "id"),
        "petstore.yaml": ("openapi", "Pet.name"),
        "user.avsc": ("avro", "name"),
        "addressbook.proto": ("protobuf", "Person.email"),
    }

    for filename, (source_format, expected_path) in expected.items():
        snapshot = inspect_file(DOWNLOADED / filename)

        assert snapshot.source_format == source_format
        assert expected_path in snapshot.field_map


def test_manual_fixture_pairs_exercise_each_supported_diff_path() -> None:
    cases = [
        (
            MANUAL / "csv/weather-baseline.csv",
            MANUAL / "csv/weather-candidate.csv",
            Severity.SAFE,
            "field_added_optional",
            "station",
        ),
        (
            MANUAL / "json/miserables-baseline.json",
            MANUAL / "json/miserables-candidate.json",
            Severity.BREAKING,
            "incompatible_type_change",
            "metadata.version",
        ),
        (
            MANUAL / "jsonl/search-baseline.jsonl",
            MANUAL / "jsonl/search-candidate.jsonl",
            Severity.BREAKING,
            "incompatible_type_change",
            "result.sequenceNumber",
        ),
        (
            MANUAL / "parquet/alltypes-baseline.parquet",
            MANUAL / "parquet/alltypes-candidate.parquet",
            Severity.SAFE,
            "field_added_optional",
            "campaign_id",
        ),
        (
            MANUAL / "openapi/petstore-baseline.yaml",
            MANUAL / "openapi/petstore-candidate.yaml",
            Severity.BREAKING,
            "incompatible_type_change",
            "Pet.name",
        ),
        (
            MANUAL / "avro/user-baseline.avsc",
            MANUAL / "avro/user-candidate.avsc",
            Severity.BREAKING,
            "incompatible_type_change",
            "favorite_number",
        ),
        (
            MANUAL / "proto/addressbook-baseline.proto",
            MANUAL / "proto/addressbook-candidate.proto",
            Severity.BREAKING,
            "incompatible_type_change",
            "Person.email",
        ),
    ]

    for old_path, new_path, severity, code, event_path in cases:
        report = compare_files(old_path, new_path, CompareOptions())

        assert report.overall == severity
        assert any(event.code == code and event.path == event_path for event in report.events)


def test_manual_option_fixtures_cover_strict_ignore_and_rename_modes() -> None:
    default_report = compare_files(
        MANUAL / "options/strict-old.csv",
        MANUAL / "options/strict-new.csv",
        CompareOptions(),
    )
    strict_report = compare_files(
        MANUAL / "options/strict-old.csv",
        MANUAL / "options/strict-new.csv",
        CompareOptions(strict=True),
    )
    rename_report = compare_files(
        MANUAL / "options/rename-old.csv",
        MANUAL / "options/rename-new.csv",
        CompareOptions(rename_heuristics=True),
    )
    ignore_report = compare_files(
        MANUAL / "options/ignore-old.jsonl",
        MANUAL / "options/ignore-new.jsonl",
        CompareOptions(ignore_fields=r"^_loaded_at$"),
    )

    assert default_report.overall == Severity.SAFE
    assert strict_report.overall == Severity.WARNING
    assert any(event.code == "possible_rename" for event in rename_report.events)
    assert ignore_report.overall == Severity.SAFE
    assert ignore_report.ignored_fields == ["_loaded_at"]


def test_manual_scan_and_baseline_workflows_use_committed_fixtures(tmp_path: Path) -> None:
    old_root = tmp_path / "old"
    new_root = tmp_path / "new"
    baseline_root = tmp_path / ".schemaglow-baseline"
    shutil.copytree(MANUAL / "scan/old", old_root)
    shutil.copytree(MANUAL / "scan/new", new_root)

    scan_report = scan_directories(old_root, new_root, CompareOptions())
    manifest = capture_baseline(old_root, baseline_root)
    baseline_report = check_baseline(baseline_root, new_root, CompareOptions())

    assert scan_report.overall == Severity.BREAKING
    assert any(entry.status == "added_file" for entry in scan_report.entries)
    assert manifest.entries
    assert baseline_report.overall == Severity.BREAKING
    assert any(entry.status == "unexpected_candidate" for entry in baseline_report.entries)



from __future__ import annotations

from pathlib import Path

from schemaglow.models import CompareOptions, Severity
from schemaglow.service import compare_files, inspect_file


def _write(path: Path, content: str) -> Path:
    path.write_text(content, encoding="utf-8")
    return path


def test_csv_diff_detects_numeric_widening_and_order_change(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "id,total\n1,10\n2,20\n")
    new_file = _write(tmp_path / "new.csv", "total,id\n10.5,1\n20.0,2\n")

    report = compare_files(old_file, new_file, options=CompareOptions())

    assert report.overall == Severity.SAFE
    assert any(event.code == "numeric_widening" for event in report.events)
    assert any(event.code == "column_order_changed" for event in report.events)


def test_jsonl_diff_detects_breaking_nullability_and_nested_expansion(tmp_path: Path) -> None:
    old_file = _write(
        tmp_path / "old.jsonl",
        "\n".join(
            [
                '{"user":{"id":1,"name":"Ada"}}',
                '{"user":{"id":2}}',
            ]
        )
        + "\n",
    )
    new_file = _write(
        tmp_path / "new.jsonl",
        "\n".join(
            [
                '{"user":{"id":1,"name":"Ada","email":"ada@example.com"}}',
                '{"user":{"id":2,"name":"Bea","email":"bea@example.com"}}',
            ]
        )
        + "\n",
    )

    report = compare_files(old_file, new_file, options=CompareOptions())

    assert any(
        event.code == "nullable_to_required" and event.path == "user.name"
        for event in report.events
    )
    assert any(
        event.code == "nested_shape_expanded" and event.path == "user.email"
        for event in report.events
    )
    assert report.overall == Severity.BREAKING


def test_rename_heuristics_flags_likely_field_rename(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "user_name\nada\nbea\n")
    new_file = _write(tmp_path / "new.csv", "username\nada\nbea\n")

    report = compare_files(old_file, new_file, options=CompareOptions(rename_heuristics=True))

    rename_events = [event for event in report.events if event.code == "possible_rename"]
    assert len(rename_events) == 1
    assert rename_events[0].summary == "possible rename: user_name -> username"


def test_inspect_file_infers_nested_json_fields(tmp_path: Path) -> None:
    path = _write(
        tmp_path / "sample.json",
        '{"order":{"id":1,"created_at":"2026-03-09T12:00:00Z"},"tags":["a","b"]}',
    )

    snapshot = inspect_file(path)
    field_map = snapshot.field_map

    assert field_map["order"].type_name == "object"
    assert field_map["order.created_at"].type_name == "datetime"
    assert field_map["tags"].type_name == "array"
    assert field_map["tags[]"].type_name == "string"


def test_openapi_avro_and_proto_are_supported(tmp_path: Path) -> None:
    openapi_file = _write(
        tmp_path / "openapi.yaml",
        """
openapi: 3.1.0
info:
  title: Demo
  version: "1.0"
paths: {}
components:
  schemas:
    User:
      type: object
      required: [id]
      properties:
        id:
          type: integer
        email:
          type: string
          format: email
""".strip()
        + "\n",
    )
    avro_file = _write(
        tmp_path / "user.avsc",
        """
{
  "type": "record",
  "name": "User",
  "fields": [
    {"name": "id", "type": "long"},
    {"name": "nickname", "type": ["null", "string"]}
  ]
}
""".strip()
        + "\n",
    )
    proto_file = _write(
        tmp_path / "user.proto",
        """
syntax = "proto3";

message User {
  string id = 1;
  repeated string tags = 2;
}
""".strip()
        + "\n",
    )

    openapi_snapshot = inspect_file(openapi_file)
    avro_snapshot = inspect_file(avro_file)
    proto_snapshot = inspect_file(proto_file)

    assert openapi_snapshot.source_format == "openapi"
    assert openapi_snapshot.field_map["User.email"].format_hint == "email"
    assert avro_snapshot.source_format == "avro"
    assert avro_snapshot.field_map["nickname"].nullable is True
    assert proto_snapshot.source_format == "protobuf"
    assert proto_snapshot.field_map["User.tags"].type_name == "array"
    assert proto_snapshot.field_map["User.tags[]"].type_name == "string"



from __future__ import annotations

from pathlib import Path

import pytest

from schemaglow.schema_sources import infer_openapi_schema, is_openapi_document
from schemaglow.service import inspect_file


def _write(path: Path, content: str) -> Path:
    path.write_text(content, encoding="utf-8")
    return path


def test_is_openapi_document_handles_supported_and_unsupported_inputs(tmp_path: Path) -> None:
    openapi_json = _write(
        tmp_path / "openapi.json",
        '{"openapi":"3.1.0","info":{"title":"Demo","version":"1.0"},"paths":{}}',
    )
    plain_json = _write(tmp_path / "plain.json", '{"value": 1}')

    assert is_openapi_document(openapi_json) is True
    assert is_openapi_document(plain_json) is False
    assert is_openapi_document(tmp_path / "notes.txt") is False


def test_openapi_paths_refs_allof_and_arrays_are_flattened(tmp_path: Path) -> None:
    path = _write(
        tmp_path / "openapi.yaml",
        """
openapi: 3.1.0
info:
  title: Demo
  version: "1.0"
paths:
  /users:
    post:
      requestBody:
        content:
          application/json:
            schema:
              $ref: '#/components/schemas/ExtendedUser'
      responses:
        "200":
          content:
            application/json:
              schema:
                type: array
                items:
                  $ref: '#/components/schemas/BaseUser'
components:
  schemas:
    BaseUser:
      type: object
      required: [id]
      properties:
        id:
          type: integer
        email:
          type: string
          format: email
    ExtendedUser:
      allOf:
        - $ref: '#/components/schemas/BaseUser'
        - type: object
          properties:
            role:
              type: string
              enum: [ADMIN, MEMBER]
""".strip()
        + "\n",
    )

    snapshot = inspect_file(path)

    assert snapshot.source_format == "openapi"
    assert snapshot.field_map["BaseUser.email"].format_hint == "email"
    assert snapshot.field_map["ExtendedUser.role"].sample_values == ["ADMIN", "MEMBER"]
    assert (
        snapshot.field_map["paths.users.post.request.application.json.role"].type_name == "string"
    )
    assert (
        snapshot.field_map["paths.users.post.responses.200.application.json"].type_name == "array"
    )
    assert (
        snapshot.field_map["paths.users.post.responses.200.application.json[].email"].format_hint
        == "email"
    )


def test_infer_openapi_rejects_non_openapi_documents(tmp_path: Path) -> None:
    path = _write(tmp_path / "plain.yaml", "name: demo\n")

    with pytest.raises(ValueError, match="does not look like an OpenAPI document"):
        infer_openapi_schema(path)


def test_avro_named_types_arrays_maps_and_unions_are_flattened(tmp_path: Path) -> None:
    path = _write(
        tmp_path / "schema.avsc",
        """
{
  "type": "record",
  "name": "User",
  "fields": [
    {"name": "profile", "type": {"type": "record", "name": "Profile", "fields": [
      {"name": "age", "type": "int"}
    ]}},
    {"name": "profile_copy", "type": "Profile"},
    {"name": "tags", "type": {"type": "array", "items": "string"}},
    {"name": "attributes", "type": {"type": "map", "values": "long"}},
    {"name": "status", "type": {"type": "enum", "name": "Status", "symbols": ["OPEN", "CLOSED"]}},
    {"name": "variant", "type": ["null", "int", "string"]}
  ]
}
""".strip()
        + "\n",
    )

    snapshot = inspect_file(path)

    assert snapshot.source_format == "avro"
    assert snapshot.field_map["profile.age"].type_name == "integer"
    assert snapshot.field_map["profile_copy.age"].type_name == "integer"
    assert snapshot.field_map["tags"].type_name == "array"
    assert snapshot.field_map["tags[]"].type_name == "string"
    assert snapshot.field_map["attributes.value"].type_name == "integer"
    assert snapshot.field_map["status"].sample_values == ["OPEN", "CLOSED"]
    assert snapshot.field_map["variant"].type_name == "mixed"
    assert snapshot.field_map["variant"].nullable is True


def test_proto_nested_messages_enums_maps_and_self_reference_are_flattened(tmp_path: Path) -> None:
    path = _write(
        tmp_path / "schema.proto",
        """
syntax = "proto3";

message Node {
  Node child = 1;
}

message User {
  enum Role {
    ROLE_UNSPECIFIED = 0;
    ADMIN = 1;
  }

  message Address {
    string city = 1;
  }

  Role role = 1;
  map<string, int64> scores = 2;
  repeated Address addresses = 3;
  optional string nickname = 4;
}

message Wrapper {
  User user = 1;
}
""".strip()
        + "\n",
    )

    snapshot = inspect_file(path)

    assert snapshot.source_format == "protobuf"
    assert snapshot.field_map["Node.child"].type_name == "object"
    assert "Node.child.child" not in snapshot.field_map
    assert snapshot.field_map["User.role"].sample_values == ["ROLE_UNSPECIFIED", "ADMIN"]
    assert snapshot.field_map["User.scores.value"].type_name == "integer"
    assert snapshot.field_map["User.addresses"].type_name == "array"
    assert snapshot.field_map["User.addresses[].city"].type_name == "string"
    assert snapshot.field_map["Wrapper.user.nickname"].nullable is True



from __future__ import annotations

from pathlib import Path

import pytest

from schemaglow.infer import detect_format
from schemaglow.models import CompareOptions, Severity
from schemaglow.service import compare_files, inspect_file


def _write(path: Path, content: str) -> Path:
    path.write_text(content, encoding="utf-8")
    return path


def test_ignore_fields_suppresses_matching_paths(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "id,_loaded_at\n1,2026-03-09\n")
    new_file = _write(tmp_path / "new.csv", "id,_loaded_at\n1,2026-03-10\n")

    report = compare_files(
        old_file, new_file, options=CompareOptions(ignore_fields=r"^_loaded_at$")
    )

    assert report.overall == Severity.SAFE
    assert report.ignored_fields == ["_loaded_at"]
    assert report.events[0].code == "no_change"


def test_strict_numeric_widening_is_warning(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "total\n10\n20\n")
    new_file = _write(tmp_path / "new.csv", "total\n10.5\n20.0\n")

    report = compare_files(old_file, new_file, options=CompareOptions(strict=True))

    assert report.overall == Severity.WARNING
    assert any(
        event.code == "numeric_widening" and event.severity == Severity.WARNING
        for event in report.events
    )


def test_required_to_nullable_becomes_warning(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "id,status\n1,open\n2,closed\n")
    new_file = _write(tmp_path / "new.csv", "id,status\n1,open\n2,\n")

    report = compare_files(old_file, new_file, options=CompareOptions())

    assert report.overall == Severity.WARNING
    assert any(event.code == "required_to_nullable" for event in report.events)


def test_top_level_json_array_of_objects_flattens_without_array_prefix(tmp_path: Path) -> None:
    source_file = _write(
        tmp_path / "records.json", '[{"user":{"id":1}},{"user":{"id":2,"name":"Bea"}}]'
    )

    snapshot = inspect_file(source_file)

    assert "user.id" in snapshot.field_map
    assert "user.name" in snapshot.field_map
    assert "[]".strip() not in snapshot.field_map


def test_detect_format_rejects_unsupported_extension() -> None:
    with pytest.raises(ValueError, match="Unsupported file format"):
        detect_format(Path("notes.txt"))


def test_sample_shape_change_is_reported_for_string_fields(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "id\nUSER_A\nUSER_B\n")
    new_file = _write(tmp_path / "new.csv", "id\nalpha-1\nbeta-2\n")

    report = compare_files(old_file, new_file, options=CompareOptions())

    assert report.overall == Severity.WARNING
    assert any(event.code == "sample_shape_changed" for event in report.events)


def test_nested_field_removal_is_collapsed_to_parent_event(tmp_path: Path) -> None:
    old_file = _write(
        tmp_path / "old.json", '{"user":{"profile":{"id":1,"email":"a@example.com"}}}'
    )
    new_file = _write(tmp_path / "new.json", '{"user":{}}')

    report = compare_files(old_file, new_file, options=CompareOptions())
    removed_paths = {event.path for event in report.events if event.code == "field_removed"}

    assert "user.profile" in removed_paths
    assert "user.profile.id" not in removed_paths


def test_type_change_branches_cover_shape_widening_ambiguous_and_materialized(
    tmp_path: Path,
) -> None:
    old_date = _write(tmp_path / "old_date.csv", "created_at\n2026-03-09\n")
    new_string = _write(tmp_path / "new_string.csv", "created_at\nnot-a-date\n")
    shape_report = compare_files(old_date, new_string, options=CompareOptions())

    old_mixed = _write(
        tmp_path / "old_mixed.jsonl",
        '{"value": 1}\n{"value": "two"}\n',
    )
    new_plain = _write(
        tmp_path / "new_plain.jsonl",
        '{"value": "one"}\n{"value": "two"}\n',
    )
    ambiguous_report = compare_files(old_mixed, new_plain, options=CompareOptions())

    old_null = _write(tmp_path / "old_null.json", '{"status": null}')
    new_value = _write(tmp_path / "new_value.json", '{"status": "open"}')
    materialized_report = compare_files(old_null, new_value, options=CompareOptions())

    old_base = _write(tmp_path / "old_base.csv", "id\n1\n")
    new_required = _write(tmp_path / "new_required.csv", "id,status\n1,open\n")
    added_required_report = compare_files(old_base, new_required, options=CompareOptions())

    assert any(event.code == "shape_widening" for event in shape_report.events)
    assert any(event.code == "ambiguous_type_change" for event in ambiguous_report.events)
    assert any(event.code == "materialized_type" for event in materialized_report.events)
    assert any(event.code == "field_added_required" for event in added_required_report.events)



from __future__ import annotations

from html import escape
from pathlib import Path
from typing import cast

from jinja2 import Template

from .models import (
    BaselineManifest,
    ComparisonReport,
    DiffEvent,
    DirectoryReport,
    SchemaSnapshot,
)

HTML_TEMPLATE = Template(
    """<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>SchemaGlow Report</title>
  <style>
    :root {
      --safe: #116149;
      --warning: #9a6700;
      --breaking: #b42318;
      --ink: #172033;
      --line: #d8deea;
      --paper: #f7f8fb;
    }
    body {
      font-family: "Iowan Old Style", "Palatino Linotype", serif;
      margin: 2rem auto;
      max-width: 960px;
      color: var(--ink);
      background: linear-gradient(180deg, #ffffff 0%, var(--paper) 100%);
      padding: 0 1rem 3rem;
    }
    h1, h2 { margin-bottom: 0.5rem; }
    .meta { color: #51607d; }
    .event {
      border: 1px solid var(--line);
      border-left-width: 6px;
      border-radius: 12px;
      background: white;
      padding: 1rem;
      margin: 0 0 1rem;
    }
    .SAFE { border-left-color: var(--safe); }
    .WARNING { border-left-color: var(--warning); }
    .BREAKING { border-left-color: var(--breaking); }
    code {
      background: #eef2f8;
      padding: 0.1rem 0.3rem;
      border-radius: 4px;
    }
  </style>
</head>
<body>
  <h1>SchemaGlow Report</h1>
  <p class="meta">
    <strong>{{ report.overall.value }}</strong>
    for {{ report.old_snapshot.source_path }} -> {{ report.new_snapshot.source_path }}
  </p>
  <p class="meta">
    Counts: safe={{ report.counts["SAFE"] }},
    warning={{ report.counts["WARNING"] }},
    breaking={{ report.counts["BREAKING"] }}
  </p>
  {% for event in report.events %}
  <div class="event {{ event.severity.value }}">
    <strong>{{ event.severity.value }}</strong>
    <p>{{ event.summary }}</p>
    {% if event.old %}
    <p>
      Before: <code>{{ event.old.type_name }}</code>
      {% if event.old.nullable %} nullable{% else %} required{% endif %}
    </p>
    {% endif %}
    {% if event.new %}
    <p>
      After: <code>{{ event.new.type_name }}</code>
      {% if event.new.nullable %} nullable{% else %} required{% endif %}
    </p>
    {% endif %}
  </div>
  {% endfor %}
</body>
</html>
"""
)

DIRECTORY_HTML_TEMPLATE = Template(
    """<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>SchemaGlow Scan Report</title>
  <style>
    body {
      font-family: "Iowan Old Style", "Palatino Linotype", serif;
      margin: 2rem auto;
      max-width: 1000px;
      color: #172033;
      background: linear-gradient(180deg, #ffffff 0%, #f7f8fb 100%);
      padding: 0 1rem 3rem;
    }
    .entry {
      border: 1px solid #d8deea;
      border-left-width: 6px;
      border-radius: 12px;
      background: white;
      padding: 1rem;
      margin: 0 0 1rem;
    }
    .SAFE { border-left-color: #116149; }
    .WARNING { border-left-color: #9a6700; }
    .BREAKING { border-left-color: #b42318; }
  </style>
</head>
<body>
  <h1>SchemaGlow Scan Report</h1>
  <p>
    <strong>{{ report.overall.value }}</strong>
    for {{ report.old_root }} -> {{ report.new_root }}
  </p>
  {% for entry in report.entries %}
  <div class="entry {{ entry.severity.value }}">
    <strong>{{ entry.severity.value }}</strong>
    <p>{{ entry.summary }}</p>
    {% if entry.report %}
      <p>
        Counts: safe={{ entry.report.counts["SAFE"] }},
        warning={{ entry.report.counts["WARNING"] }},
        breaking={{ entry.report.counts["BREAKING"] }}
      </p>
    {% endif %}
  </div>
  {% endfor %}
</body>
</html>
"""
)


def render_text_report(report: ComparisonReport) -> str:
    lines = [
        "SchemaGlow Report",
        "",
        f"{report.overall.value}",
        f"old: {report.old_snapshot.source_path}",
        f"new: {report.new_snapshot.source_path}",
        (
            "counts: "
            f"SAFE={report.counts['SAFE']} "
            f"WARNING={report.counts['WARNING']} "
            f"BREAKING={report.counts['BREAKING']}"
        ),
    ]
    if report.ignored_fields:
        lines.append(f"ignored: {', '.join(report.ignored_fields)}")
    for event in report.events:
        lines.extend(_format_event_lines(event))
    return "\n".join(lines)


def render_json_report(report: ComparisonReport) -> str:
    try:
        import orjson  # type: ignore[import-not-found]

        payload = cast(
            bytes,
            orjson.dumps(_report_payload(report), option=orjson.OPT_INDENT_2),
        )
        return payload.decode("utf-8")
    except ModuleNotFoundError:
        import json

        return json.dumps(_report_payload(report), indent=2)


def render_markdown_report(report: ComparisonReport) -> str:
    lines = [
        "# SchemaGlow Report",
        "",
        f"**Overall:** `{report.overall.value}`",
        "",
        f"- Old snapshot: `{report.old_snapshot.source_path}`",
        f"- New snapshot: `{report.new_snapshot.source_path}`",
        (
            "- Counts: "
            f"SAFE={report.counts['SAFE']}, "
            f"WARNING={report.counts['WARNING']}, "
            f"BREAKING={report.counts['BREAKING']}"
        ),
        "",
    ]
    if report.ignored_fields:
        ignored_lines = [f"- `{field}`" for field in report.ignored_fields]
        lines.extend(["## Ignored Fields", "", *ignored_lines, ""])
    lines.extend(["## Events", ""])
    for event in report.events:
        lines.append(f"### {event.severity.value}")
        lines.append(f"- Path: `{event.path}`")
        lines.append(f"- Summary: {event.summary}")
        if event.old is not None:
            lines.append(
                "- Before: "
                f"`{event.old.type_name}` / "
                f"{'nullable' if event.old.nullable else 'required'}"
            )
        if event.new is not None:
            lines.append(
                "- After: "
                f"`{event.new.type_name}` / "
                f"{'nullable' if event.new.nullable else 'required'}"
            )
        lines.append("")
    return "\n".join(lines)


def render_html_report(report: ComparisonReport) -> str:
    return HTML_TEMPLATE.render(report=report)


def render_directory_text(report: DirectoryReport) -> str:
    lines = [
        "SchemaGlow Scan Report",
        "",
        f"{report.overall.value}",
        f"old: {report.old_root}",
        f"new: {report.new_root}",
        (
            "counts: "
            f"SAFE={report.counts['SAFE']} "
            f"WARNING={report.counts['WARNING']} "
            f"BREAKING={report.counts['BREAKING']}"
        ),
        "",
    ]
    for entry in report.entries:
        lines.append(f"{entry.severity.value} {entry.relative_path}")
        lines.append(f"- {entry.summary}")
        if entry.report is not None:
            for event in entry.report.events:
                lines.extend(_format_event_lines(event))
        lines.append("")
    return "\n".join(lines).strip()


def render_directory_json(report: DirectoryReport) -> str:
    try:
        import orjson

        payload = cast(
            bytes,
            orjson.dumps(_directory_payload(report), option=orjson.OPT_INDENT_2),
        )
        return payload.decode("utf-8")
    except ModuleNotFoundError:
        import json

        return json.dumps(_directory_payload(report), indent=2)


def render_directory_markdown(report: DirectoryReport) -> str:
    lines = [
        "# SchemaGlow Scan Report",
        "",
        f"**Overall:** `{report.overall.value}`",
        "",
        f"- Old root: `{report.old_root}`",
        f"- New root: `{report.new_root}`",
        (
            "- Counts: "
            f"SAFE={report.counts['SAFE']}, "
            f"WARNING={report.counts['WARNING']}, "
            f"BREAKING={report.counts['BREAKING']}"
        ),
        "",
    ]
    for entry in report.entries:
        lines.append(f"## {entry.relative_path}")
        lines.append(f"- Severity: `{entry.severity.value}`")
        lines.append(f"- Summary: {entry.summary}")
        if entry.report is not None:
            for event in entry.report.events:
                lines.append(f"- Event: {event.summary}")
        lines.append("")
    return "\n".join(lines)


def render_directory_html(report: DirectoryReport) -> str:
    return DIRECTORY_HTML_TEMPLATE.render(report=report)


def render_snapshot_text(snapshot: SchemaSnapshot) -> str:
    lines = [
        f"SchemaGlow Inspect: {snapshot.source_path}",
        f"format: {snapshot.source_format}",
        f"root: {snapshot.root_type}",
        f"fields: {len(snapshot.fields)}",
        "",
    ]
    for field in snapshot.fields:
        nullability = "nullable" if field.nullable else "required"
        sample = f" samples={field.sample_values}" if field.sample_values else ""
        lines.append(f"- {field.path}: {field.type_name} ({nullability}){sample}")
    return "\n".join(lines)


def write_report(path: Path, content: str) -> None:
    path.write_text(content, encoding="utf-8")


def render_baseline_manifest_text(manifest: BaselineManifest, baseline_root: Path) -> str:
    lines = [
        "SchemaGlow Baseline",
        "",
        f"source: {manifest.source_root}",
        f"baseline: {baseline_root}",
        f"files: {len(manifest.entries)}",
        f"sample_size: {manifest.sample_size}",
        "",
    ]
    for entry in manifest.entries:
        lines.append(f"- {entry.relative_path} -> {entry.snapshot_path} ({entry.source_format})")
    return "\n".join(lines)


def _report_payload(report: ComparisonReport) -> dict[str, object]:
    payload = report.model_dump(mode="json")
    payload["overall"] = report.overall.value
    payload["counts"] = report.counts
    return payload


def _directory_payload(report: DirectoryReport) -> dict[str, object]:
    payload = report.model_dump(mode="json")
    payload["overall"] = report.overall.value
    payload["counts"] = report.counts
    for entry_payload, entry in zip(payload["entries"], report.entries, strict=False):
        if isinstance(entry_payload, dict) and entry.report is not None:
            entry_payload["report"] = _report_payload(entry.report)
    return payload


def _format_event_lines(event: DiffEvent) -> list[str]:
    prefix = {
        "SAFE": "+",
        "WARNING": "~",
        "BREAKING": "-",
    }[event.severity.value]
    lines = [f"{event.severity.value}", f"{prefix} {event.summary}"]
    if event.old is not None:
        lines.append(
            "  before: "
            f"{escape(event.old.type_name)} / "
            f"{'nullable' if event.old.nullable else 'required'}"
        )
    if event.new is not None:
        lines.append(
            "  after: "
            f"{escape(event.new.type_name)} / "
            f"{'nullable' if event.new.nullable else 'required'}"
        )
    return lines


__author__ = "gadwant"



from __future__ import annotations

from datetime import UTC, datetime
from enum import StrEnum

from pydantic import BaseModel, Field


class Severity(StrEnum):
    SAFE = "SAFE"
    WARNING = "WARNING"
    BREAKING = "BREAKING"


class FieldSchema(BaseModel):
    path: str
    type_name: str
    nullable: bool = False
    format_hint: str | None = None
    position: int | None = None
    sample_values: list[str] = Field(default_factory=list)


class SchemaSnapshot(BaseModel):
    source_path: str | None = None
    source_format: str
    root_type: str
    sample_size: int
    generated_at: datetime = Field(default_factory=lambda: datetime.now(UTC))
    fields: list[FieldSchema]
    metadata: dict[str, str | int | bool | list[str]] = Field(default_factory=dict)

    @property
    def field_map(self) -> dict[str, FieldSchema]:
        return {field.path: field for field in self.fields}


class DiffEvent(BaseModel):
    severity: Severity
    code: str
    path: str
    summary: str
    old: FieldSchema | None = None
    new: FieldSchema | None = None


class ComparisonReport(BaseModel):
    old_snapshot: SchemaSnapshot
    new_snapshot: SchemaSnapshot
    events: list[DiffEvent]
    ignored_fields: list[str] = Field(default_factory=list)

    @property
    def overall(self) -> Severity:
        if any(event.severity == Severity.BREAKING for event in self.events):
            return Severity.BREAKING
        if any(event.severity == Severity.WARNING for event in self.events):
            return Severity.WARNING
        return Severity.SAFE

    @property
    def counts(self) -> dict[str, int]:
        counts = {severity.value: 0 for severity in Severity}
        for event in self.events:
            counts[event.severity.value] += 1
        return counts


class DirectoryEntry(BaseModel):
    relative_path: str
    severity: Severity
    status: str
    summary: str
    report: ComparisonReport | None = None


class DirectoryReport(BaseModel):
    old_root: str
    new_root: str
    entries: list[DirectoryEntry]

    @property
    def overall(self) -> Severity:
        if any(entry.severity == Severity.BREAKING for entry in self.entries):
            return Severity.BREAKING
        if any(entry.severity == Severity.WARNING for entry in self.entries):
            return Severity.WARNING
        return Severity.SAFE

    @property
    def counts(self) -> dict[str, int]:
        counts = {severity.value: 0 for severity in Severity}
        for entry in self.entries:
            counts[entry.severity.value] += 1
        return counts


class BaselineEntry(BaseModel):
    relative_path: str
    snapshot_path: str
    source_format: str


class BaselineManifest(BaseModel):
    source_root: str
    generated_at: datetime = Field(default_factory=lambda: datetime.now(UTC))
    sample_size: int
    entries: list[BaselineEntry]


class CompareOptions(BaseModel):
    sample_size: int = 10_000
    ignore_order: bool = False
    ignore_fields: str | None = None
    strict: bool = False
    rename_heuristics: bool = False


__author__ = "gadwant"



from __future__ import annotations

import re
from difflib import SequenceMatcher

from .models import (
    CompareOptions,
    ComparisonReport,
    DiffEvent,
    FieldSchema,
    SchemaSnapshot,
    Severity,
)


def compare_schema_snapshots(
    old_snapshot: SchemaSnapshot, new_snapshot: SchemaSnapshot, options: CompareOptions
) -> ComparisonReport:
    old_fields, ignored = _filter_fields(old_snapshot.fields, options.ignore_fields)
    new_fields, ignored_new = _filter_fields(new_snapshot.fields, options.ignore_fields)
    ignored.extend(ignored_new)
    old_map = {field.path: field for field in old_fields}
    new_map = {field.path: field for field in new_fields}
    old_paths = set(old_map)
    new_paths = set(new_map)
    common_paths = old_paths & new_paths
    removed = set(old_paths - new_paths)
    added = set(new_paths - old_paths)
    events: list[DiffEvent] = []

    removed = _collapse_nested_paths(removed)
    added = _collapse_nested_paths(added)

    if options.rename_heuristics:
        rename_events, removed, added = _detect_possible_renames(removed, added, old_map, new_map)
        events.extend(rename_events)

    for path in sorted(common_paths):
        old_field = old_map[path]
        new_field = new_map[path]
        events.extend(_compare_common_field(path, old_field, new_field, options.strict))

    for path in sorted(removed):
        events.append(
            DiffEvent(
                severity=Severity.BREAKING,
                code="field_removed",
                path=path,
                summary=f"removed field: {path}",
                old=old_map[path],
            )
        )

    for path in sorted(added):
        new_field = new_map[path]
        severity, code, summary = _classify_added_field(path, new_field)
        events.append(
            DiffEvent(
                severity=severity,
                code=code,
                path=path,
                summary=summary,
                new=new_field,
            )
        )

    if not options.ignore_order:
        order_event = _detect_order_change(old_fields, new_fields)
        if order_event is not None:
            events.append(order_event)

    if not events:
        events.append(
            DiffEvent(
                severity=Severity.SAFE,
                code="no_change",
                path="*",
                summary="no schema changes detected",
            )
        )

    return ComparisonReport(
        old_snapshot=old_snapshot.model_copy(update={"fields": old_fields}),
        new_snapshot=new_snapshot.model_copy(update={"fields": new_fields}),
        events=sorted(events, key=_event_sort_key),
        ignored_fields=sorted(set(ignored)),
    )


def _filter_fields(
    fields: list[FieldSchema], pattern: str | None
) -> tuple[list[FieldSchema], list[str]]:
    if pattern is None:
        return fields, []
    regex = re.compile(pattern)
    kept: list[FieldSchema] = []
    ignored: list[str] = []
    for field in fields:
        if regex.search(field.path) or _has_ignored_ancestor(field.path, ignored):
            ignored.append(field.path)
            continue
        kept.append(field)
    return kept, ignored


def _compare_common_field(
    path: str, old_field: FieldSchema, new_field: FieldSchema, strict: bool
) -> list[DiffEvent]:
    events: list[DiffEvent] = []
    if old_field.type_name != new_field.type_name:
        severity, code = _classify_type_change(old_field.type_name, new_field.type_name, strict)
        events.append(
            DiffEvent(
                severity=severity,
                code=code,
                path=path,
                summary=f"type changed: {path} ({old_field.type_name} -> {new_field.type_name})",
                old=old_field,
                new=new_field,
            )
        )
    if old_field.nullable and not new_field.nullable:
        events.append(
            DiffEvent(
                severity=Severity.BREAKING,
                code="nullable_to_required",
                path=path,
                summary=f"nullability changed: {path} (nullable -> required)",
                old=old_field,
                new=new_field,
            )
        )
    elif not old_field.nullable and new_field.nullable:
        events.append(
            DiffEvent(
                severity=Severity.WARNING,
                code="required_to_nullable",
                path=path,
                summary=f"nullability changed: {path} (required -> nullable)",
                old=old_field,
                new=new_field,
            )
        )
    shape_change = _detect_sample_shape_change(path, old_field, new_field)
    if shape_change is not None:
        events.append(shape_change)
    return events


def _classify_type_change(old_type: str, new_type: str, strict: bool) -> tuple[Severity, str]:
    if old_type == "integer" and new_type == "number":
        return (
            (Severity.WARNING, "numeric_widening")
            if strict
            else (Severity.SAFE, "numeric_widening")
        )
    if old_type in {"date", "datetime"} and new_type == "string":
        return Severity.WARNING, "shape_widening"
    if old_type == "mixed" or new_type == "mixed":
        return Severity.WARNING, "ambiguous_type_change"
    if old_type == new_type:
        return Severity.SAFE, "no_change"
    if old_type == "null":
        return Severity.SAFE, "materialized_type"
    return Severity.BREAKING, "incompatible_type_change"


def _classify_added_field(path: str, field: FieldSchema) -> tuple[Severity, str, str]:
    nested = "." in path or "[]" in path
    if nested:
        return Severity.WARNING, "nested_shape_expanded", f"nested object shape expanded: {path}"
    if field.nullable:
        return (
            Severity.SAFE,
            "field_added_optional",
            f"added optional field: {path} ({field.type_name})",
        )
    return (
        Severity.WARNING,
        "field_added_required",
        f"added required field: {path} ({field.type_name})",
    )


def _detect_possible_renames(
    removed: set[str],
    added: set[str],
    old_map: dict[str, FieldSchema],
    new_map: dict[str, FieldSchema],
) -> tuple[list[DiffEvent], set[str], set[str]]:
    events: list[DiffEvent] = []
    matched_removed: set[str] = set()
    matched_added: set[str] = set()
    for removed_path in sorted(removed):
        old_parent = removed_path.rpartition(".")[0]
        best_match: str | None = None
        best_score = 0.0
        for added_path in sorted(added):
            if added_path in matched_added:
                continue
            if added_path.rpartition(".")[0] != old_parent:
                continue
            if old_map[removed_path].type_name != new_map[added_path].type_name:
                continue
            if old_map[removed_path].nullable != new_map[added_path].nullable:
                continue
            score = SequenceMatcher(
                None, removed_path.split(".")[-1], added_path.split(".")[-1]
            ).ratio()
            sample_overlap = _sample_overlap_score(
                old_map[removed_path].sample_values,
                new_map[added_path].sample_values,
            )
            score = (score * 0.8) + (sample_overlap * 0.2)
            if score > best_score:
                best_match = added_path
                best_score = score
        if best_match is not None and best_score >= 0.72:
            matched_removed.add(removed_path)
            matched_added.add(best_match)
            events.append(
                DiffEvent(
                    severity=Severity.WARNING,
                    code="possible_rename",
                    path=removed_path,
                    summary=f"possible rename: {removed_path} -> {best_match}",
                    old=old_map[removed_path],
                    new=new_map[best_match],
                )
            )
    return events, removed - matched_removed, added - matched_added


def _detect_order_change(
    old_fields: list[FieldSchema], new_fields: list[FieldSchema]
) -> DiffEvent | None:
    old_order = [
        field.path for field in old_fields if "." not in field.path and "[]" not in field.path
    ]
    new_order = [
        field.path for field in new_fields if "." not in field.path and "[]" not in field.path
    ]
    if old_order and old_order != new_order and set(old_order) == set(new_order):
        return DiffEvent(
            severity=Severity.SAFE,
            code="column_order_changed",
            path="*",
            summary="column order changed only",
        )
    return None


def _event_sort_key(event: DiffEvent) -> tuple[int, str, str]:
    rank = {
        Severity.BREAKING: 0,
        Severity.WARNING: 1,
        Severity.SAFE: 2,
    }[event.severity]
    return rank, event.path, event.code


def _collapse_nested_paths(paths: set[str]) -> set[str]:
    collapsed: set[str] = set()
    for path in sorted(paths, key=lambda item: (item.count("."), item.count("[]"), item)):
        if any(_is_descendant(path, ancestor) for ancestor in collapsed):
            continue
        collapsed.add(path)
    return collapsed


def _has_ignored_ancestor(path: str, ignored: list[str]) -> bool:
    return any(_is_descendant(path, ancestor) for ancestor in ignored)


def _is_descendant(path: str, ancestor: str) -> bool:
    return path.startswith(f"{ancestor}.") or path.startswith(f"{ancestor}[]")


def _detect_sample_shape_change(
    path: str, old_field: FieldSchema, new_field: FieldSchema
) -> DiffEvent | None:
    if old_field.type_name != "string" or new_field.type_name != "string":
        return None
    old_shape = _sample_shape(old_field.sample_values)
    new_shape = _sample_shape(new_field.sample_values)
    if old_shape in {"unknown", "mixed"} or new_shape in {"unknown", "mixed"}:
        return None
    if old_shape == new_shape:
        return None
    return DiffEvent(
        severity=Severity.WARNING,
        code="sample_shape_changed",
        path=path,
        summary=f"sample shape changed: {path} ({old_shape} -> {new_shape})",
        old=old_field,
        new=new_field,
    )


def _sample_shape(values: list[str]) -> str:
    if not values:
        return "unknown"
    shapes = {_classify_sample_value(value) for value in values}
    if len(shapes) == 1:
        return next(iter(shapes))
    return "mixed"


def _classify_sample_value(value: str) -> str:
    stripped = value.strip()
    if re.fullmatch(r"-?\d+", stripped):
        return "integer-like"
    if re.fullmatch(r"-?(?:\d+\.\d+|\d+\.\d*|\.\d+)", stripped):
        return "number-like"
    if re.fullmatch(r"[0-9a-fA-F-]{36}", stripped):
        return "uuid-like"
    if re.fullmatch(r"[A-Z0-9_]+", stripped):
        return "enum-like"
    if re.fullmatch(r"[A-Za-z_][A-Za-z0-9_-]*", stripped):
        return "identifier-like"
    if " " in stripped:
        return "free-text"
    return "string-like"


def _sample_overlap_score(old_samples: list[str], new_samples: list[str]) -> float:
    if not old_samples or not new_samples:
        return 0.0
    old_set = set(old_samples)
    new_set = set(new_samples)
    return len(old_set & new_set) / max(len(old_set | new_set), 1)


__author__ = "gadwant"



from __future__ import annotations

import csv
import json
import re
from collections.abc import Iterable
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

import pyarrow.parquet as pq
import pyarrow.types as pat

from .models import FieldSchema, SchemaSnapshot
from .schema_sources import (
    infer_avro_schema,
    infer_openapi_schema,
    infer_proto_schema,
    is_openapi_document,
)

INTEGER_RE = re.compile(r"^-?\d+$")
FLOAT_RE = re.compile(r"^-?(?:\d+\.\d+|\d+\.\d*|\.\d+)$")
DATE_RE = re.compile(r"^\d{4}-\d{2}-\d{2}$")
DATETIME_RE = re.compile(
    r"^\d{4}-\d{2}-\d{2}[tT ][0-2]\d:[0-5]\d"
    r"(?::[0-5]\d(?:\.\d+)?)?(?:Z|[+-][0-2]\d:[0-5]\d)?$"
)


@dataclass
class NodeAccumulator:
    observed_count: int = 0
    null_count: int = 0
    kinds: set[str] = field(default_factory=set)
    sample_values: list[str] = field(default_factory=list)
    children: dict[str, NodeAccumulator] = field(default_factory=dict)
    item: NodeAccumulator | None = None
    order: list[str] = field(default_factory=list)

    @property
    def non_null_count(self) -> int:
        return self.observed_count - self.null_count


def detect_format(path: Path) -> str:
    suffix = path.suffix.lower()
    if suffix == ".csv":
        return "csv"
    if suffix == ".json":
        if is_openapi_document(path):
            return "openapi"
        return "json"
    if suffix == ".jsonl":
        return "jsonl"
    if suffix == ".parquet":
        return "parquet"
    if suffix == ".avsc":
        return "avro"
    if suffix == ".proto":
        return "protobuf"
    if suffix in {".yaml", ".yml"} and is_openapi_document(path):
        return "openapi"
    raise ValueError(f"Unsupported file format for {path}")


def infer_schema(path: Path, sample_size: int = 10_000) -> SchemaSnapshot:
    file_format = detect_format(path)
    if file_format == "csv":
        return _infer_csv(path, sample_size)
    if file_format == "json":
        return _infer_json(path, sample_size)
    if file_format == "jsonl":
        return _infer_jsonl(path, sample_size)
    if file_format == "parquet":
        return _infer_parquet(path)
    if file_format == "openapi":
        return infer_openapi_schema(path)
    if file_format == "avro":
        return infer_avro_schema(path)
    if file_format == "protobuf":
        return infer_proto_schema(path)
    raise AssertionError("unreachable")


def _infer_csv(path: Path, sample_size: int) -> SchemaSnapshot:
    with path.open("r", encoding="utf-8", newline="") as handle:
        reader = csv.DictReader(handle)
        if reader.fieldnames is None:
            raise ValueError(f"CSV file {path} is missing a header row")
        accumulators = {name: NodeAccumulator() for name in reader.fieldnames}
        row_count = 0
        for row in reader:
            row_count += 1
            for position, column in enumerate(reader.fieldnames):
                value = row.get(column)
                node = accumulators[column]
                node.observed_count += 1
                if value is None or value == "":
                    node.null_count += 1
                    node.kinds.add("null")
                    continue
                node.kinds.add(_infer_scalar_type(value))
                _append_sample(node.sample_values, value)
                if node.order == []:
                    node.order = [str(position)]
            if row_count >= sample_size:
                break
    fields = [
        FieldSchema(
            path=column,
            type_name=_normalize_kinds(node.kinds),
            nullable=node.null_count > 0,
            position=index,
            sample_values=node.sample_values,
        )
        for index, (column, node) in enumerate(accumulators.items())
    ]
    return SchemaSnapshot(
        source_path=str(path),
        source_format="csv",
        root_type="object",
        sample_size=row_count,
        fields=fields,
        metadata={"columns": list(reader.fieldnames), "rows_sampled": row_count},
    )


def _load_json(path: Path) -> Any:
    with path.open("rb") as handle:
        raw = handle.read()
    try:
        import orjson  # type: ignore

        return orjson.loads(raw)
    except ModuleNotFoundError:
        return json.loads(raw.decode("utf-8"))


def _infer_json(path: Path, sample_size: int) -> SchemaSnapshot:
    payload = _load_json(path)
    return _infer_structured_payload(path, "json", payload, sample_size)


def _infer_jsonl(path: Path, sample_size: int) -> SchemaSnapshot:
    rows: list[Any] = []
    with path.open("r", encoding="utf-8") as handle:
        for line in handle:
            stripped = line.strip()
            if not stripped:
                continue
            rows.append(json.loads(stripped))
            if len(rows) >= sample_size:
                break
    return _infer_structured_payload(path, "jsonl", rows, sample_size)


def _infer_structured_payload(
    path: Path, file_format: str, payload: Any, sample_size: int
) -> SchemaSnapshot:
    root = NodeAccumulator()
    top_level = payload
    sampled = 1
    if isinstance(payload, list):
        top_level = payload[:sample_size]
        sampled = len(top_level)
    observe_value(root, top_level)
    if isinstance(top_level, list) and root.item is not None and "object" in root.item.kinds:
        fields = flatten_node(root.item, "", parent_count=max(sampled, 1))
    else:
        parent_count = 1 if not isinstance(top_level, list) else max(sampled, 1)
        fields = flatten_node(root, "", parent_count=parent_count)
    return SchemaSnapshot(
        source_path=str(path),
        source_format=file_format,
        root_type=_normalize_kinds(root.kinds),
        sample_size=sampled,
        fields=fields,
        metadata={"rows_sampled": sampled},
    )


def _infer_parquet(path: Path) -> SchemaSnapshot:
    schema = pq.read_schema(path)
    fields = _flatten_arrow_schema(schema)
    return SchemaSnapshot(
        source_path=str(path),
        source_format="parquet",
        root_type="object",
        sample_size=schema.names.__len__(),
        fields=fields,
        metadata={"columns": list(schema.names)},
    )


def observe_value(node: NodeAccumulator, value: Any) -> None:
    node.observed_count += 1
    if value is None:
        node.null_count += 1
        node.kinds.add("null")
        return

    if isinstance(value, bool):
        node.kinds.add("boolean")
        _append_sample(node.sample_values, str(value).lower())
        return
    if isinstance(value, int) and not isinstance(value, bool):
        node.kinds.add("integer")
        _append_sample(node.sample_values, str(value))
        return
    if isinstance(value, float):
        node.kinds.add("number")
        _append_sample(node.sample_values, str(value))
        return
    if isinstance(value, str):
        node.kinds.add(_infer_scalar_type(value))
        _append_sample(node.sample_values, value)
        return
    if isinstance(value, dict):
        node.kinds.add("object")
        for key, child_value in value.items():
            if key not in node.children:
                node.children[key] = NodeAccumulator()
                node.order.append(key)
            observe_value(node.children[key], child_value)
        return
    if isinstance(value, list):
        node.kinds.add("array")
        if node.item is None:
            node.item = NodeAccumulator()
        for item in value:
            observe_value(node.item, item)
        return

    node.kinds.add("string")
    _append_sample(node.sample_values, str(value))


def flatten_node(node: NodeAccumulator, path: str, parent_count: int) -> list[FieldSchema]:
    fields: list[FieldSchema] = []
    non_null_kinds = node.kinds - {"null"}
    effective_type = _normalize_kinds(node.kinds)
    if path:
        nullable = node.null_count > 0 or node.observed_count < parent_count
        fields.append(
            FieldSchema(
                path=path,
                type_name=effective_type,
                nullable=nullable,
                sample_values=node.sample_values,
            )
        )

    if "object" in non_null_kinds:
        for index, key in enumerate(node.order):
            child = node.children[key]
            child_path = key if not path else f"{path}.{key}"
            child_fields = flatten_node(child, child_path, max(node.non_null_count, 1))
            if child_fields:
                child_fields[0].position = index
            fields.extend(child_fields)

    if "array" in non_null_kinds and node.item is not None:
        item_path = f"{path}[]" if path else "[]"
        fields.extend(flatten_node(node.item, item_path, max(node.item.observed_count, 1)))
    return fields


def _flatten_arrow_schema(schema: Any) -> list[FieldSchema]:
    fields: list[FieldSchema] = []
    for index, schema_field in enumerate(schema):
        fields.extend(_flatten_arrow_field(schema_field, schema_field.name, position=index))
    return fields


def _flatten_arrow_field(field: Any, path: str, position: int | None = None) -> list[FieldSchema]:
    fields = [
        FieldSchema(
            path=path,
            type_name=_arrow_type_name(field.type),
            nullable=field.nullable,
            position=position,
        )
    ]
    if pat.is_struct(field.type):
        for index, child in enumerate(field.type):
            fields.extend(_flatten_arrow_field(child, f"{path}.{child.name}", position=index))
    elif pat.is_list(field.type) or pat.is_large_list(field.type):
        item_field = field.type.value_field
        fields.extend(_flatten_arrow_field(item_field, f"{path}[]"))
    return fields


def _arrow_type_name(arrow_type: Any) -> str:
    if pat.is_boolean(arrow_type):
        return "boolean"
    if pat.is_integer(arrow_type):
        return "integer"
    if pat.is_floating(arrow_type) or pat.is_decimal(arrow_type):
        return "number"
    if pat.is_date(arrow_type):
        return "date"
    if pat.is_timestamp(arrow_type) or pat.is_time(arrow_type):
        return "datetime"
    if pat.is_string(arrow_type) or pat.is_large_string(arrow_type):
        return "string"
    if pat.is_binary(arrow_type) or pat.is_large_binary(arrow_type):
        return "binary"
    if pat.is_struct(arrow_type):
        return "object"
    if pat.is_list(arrow_type) or pat.is_large_list(arrow_type):
        return "array"
    return "string"


def _infer_scalar_type(value: str) -> str:
    normalized = value.strip()
    lowered = normalized.lower()
    if lowered in {"true", "false"}:
        return "boolean"
    if INTEGER_RE.match(normalized):
        return "integer"
    if FLOAT_RE.match(normalized):
        return "number"
    if DATE_RE.match(normalized):
        return "date"
    if DATETIME_RE.match(normalized):
        return "datetime"
    return "string"


def _normalize_kinds(kinds: Iterable[str]) -> str:
    non_null = {kind for kind in kinds if kind != "null"}
    if not non_null:
        return "null"
    if len(non_null) == 1:
        return next(iter(non_null))
    if non_null <= {"integer", "number"}:
        return "number"
    return "mixed"


def _append_sample(samples: list[str], value: str, limit: int = 3) -> None:
    if value not in samples and len(samples) < limit:
        samples.append(value)


__author__ = "gadwant"



from __future__ import annotations

import json
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from .models import FieldSchema, SchemaSnapshot

HTTP_METHODS = {"get", "put", "post", "delete", "patch", "head", "options", "trace"}
PROTO_FIELD_RE = re.compile(r"(?:(optional|required|repeated)\s+)?([.\w]+)\s+(\w+)\s*=\s*\d+")
PROTO_MAP_RE = re.compile(
    r"(?:(optional|required|repeated)\s+)?map<\s*([^,>]+)\s*,\s*([^>]+)\s*>\s+(\w+)\s*=\s*\d+"
)


def is_openapi_document(path: Path) -> bool:
    try:
        payload = _load_structured_document(path)
    except ValueError:
        return False
    return _looks_like_openapi(payload)


def infer_openapi_schema(path: Path) -> SchemaSnapshot:
    payload = _load_structured_document(path)
    if not _looks_like_openapi(payload):
        raise ValueError(f"{path} does not look like an OpenAPI document")

    resolver = _OpenApiResolver(payload)
    fields: list[FieldSchema] = []

    components = payload.get("components", {}).get("schemas", {})
    for index, (name, schema) in enumerate(components.items()):
        component_fields = _flatten_openapi_schema(
            schema,
            name,
            resolver=resolver,
            required=True,
            ref_chain=(),
        )
        if component_fields:
            component_fields[0].position = index
        fields.extend(component_fields)

    paths = payload.get("paths", {})
    for path_name, path_item in paths.items():
        if not isinstance(path_item, dict):
            continue
        safe_path = _safe_segment(path_name)
        for method, operation in path_item.items():
            if method.lower() not in HTTP_METHODS or not isinstance(operation, dict):
                continue
            prefix = f"paths.{safe_path}.{method.lower()}"
            request_body = operation.get("requestBody", {})
            if isinstance(request_body, dict):
                content = request_body.get("content", {})
                fields.extend(_flatten_openapi_content(content, f"{prefix}.request", resolver))
            responses = operation.get("responses", {})
            if isinstance(responses, dict):
                for status_code, response in responses.items():
                    if not isinstance(response, dict):
                        continue
                    content = response.get("content", {})
                    response_prefix = f"{prefix}.responses.{_safe_segment(str(status_code))}"
                    fields.extend(_flatten_openapi_content(content, response_prefix, resolver))

    return SchemaSnapshot(
        source_path=str(path),
        source_format="openapi",
        root_type="object",
        sample_size=len(fields),
        fields=fields,
        metadata={"schemas": list(components), "paths": len(paths)},
    )


def infer_avro_schema(path: Path) -> SchemaSnapshot:
    payload = json.loads(path.read_text(encoding="utf-8"))
    named_types: dict[str, Any] = {}
    _register_avro_types(payload, named_types)
    fields = _flatten_avro_schema(payload, "", named_types=named_types, required=True)
    return SchemaSnapshot(
        source_path=str(path),
        source_format="avro",
        root_type="object",
        sample_size=len(fields),
        fields=fields,
        metadata={"named_types": sorted(named_types)},
    )


def infer_proto_schema(path: Path) -> SchemaSnapshot:
    text = path.read_text(encoding="utf-8")
    messages = _parse_proto(text)
    message_lookup = {message.name: message for message in messages}
    enum_lookup = _collect_proto_enums(messages)
    fields: list[FieldSchema] = []
    for index, message in enumerate(messages):
        message_fields = _flatten_proto_message(
            message,
            path=message.name,
            message_lookup=message_lookup,
            enum_lookup=enum_lookup,
            stack=(),
        )
        if message_fields:
            message_fields[0].position = index
        fields.extend(message_fields)
    return SchemaSnapshot(
        source_path=str(path),
        source_format="protobuf",
        root_type="object",
        sample_size=len(messages),
        fields=fields,
        metadata={"messages": [message.name for message in messages]},
    )


def _load_structured_document(path: Path) -> Any:
    if path.suffix.lower() == ".json":
        return json.loads(path.read_text(encoding="utf-8"))
    if path.suffix.lower() in {".yaml", ".yml"}:
        try:
            import yaml
        except ModuleNotFoundError as exc:  # pragma: no cover
            raise ValueError("PyYAML is required for YAML OpenAPI documents") from exc
        return yaml.safe_load(path.read_text(encoding="utf-8"))
    raise ValueError(f"Unsupported structured document: {path}")


def _looks_like_openapi(payload: Any) -> bool:
    if not isinstance(payload, dict):
        return False
    return ("openapi" in payload or "swagger" in payload) and (
        "paths" in payload or payload.get("components", {}).get("schemas")
    )


def _flatten_openapi_content(
    content: Any, prefix: str, resolver: _OpenApiResolver
) -> list[FieldSchema]:
    fields: list[FieldSchema] = []
    if not isinstance(content, dict):
        return fields
    for media_type, media_payload in content.items():
        if not isinstance(media_payload, dict) or "schema" not in media_payload:
            continue
        media_prefix = f"{prefix}.{_safe_segment(media_type)}"
        fields.extend(
            _flatten_openapi_schema(
                media_payload["schema"],
                media_prefix,
                resolver=resolver,
                required=True,
                ref_chain=(),
            )
        )
    return fields


def _flatten_openapi_schema(
    schema: Any,
    path: str,
    resolver: _OpenApiResolver,
    required: bool,
    ref_chain: tuple[str, ...],
) -> list[FieldSchema]:
    if not isinstance(schema, dict):
        return [FieldSchema(path=path, type_name="string", nullable=not required)]

    if "$ref" in schema:
        ref = schema["$ref"]
        if ref in ref_chain:
            return [FieldSchema(path=path, type_name="object", nullable=not required)]
        return _flatten_openapi_schema(
            resolver.resolve_ref(ref),
            path,
            resolver=resolver,
            required=required,
            ref_chain=(*ref_chain, ref),
        )

    if "allOf" in schema:
        merged = _merge_openapi_all_of(schema["allOf"], resolver)
        schema = {**schema, **merged}

    nullable = bool(schema.get("nullable")) or not required
    sample_values = [str(value) for value in schema.get("enum", [])[:3]]
    type_name = _openapi_type_name(schema)
    format_hint = schema.get("format")
    fields = [
        FieldSchema(
            path=path,
            type_name=type_name,
            nullable=nullable,
            format_hint=format_hint,
            sample_values=sample_values,
        )
    ]

    if type_name == "object":
        properties = schema.get("properties", {})
        required_children = set(schema.get("required", []))
        if isinstance(properties, dict):
            for index, (child_name, child_schema) in enumerate(properties.items()):
                child_fields = _flatten_openapi_schema(
                    child_schema,
                    f"{path}.{child_name}",
                    resolver=resolver,
                    required=child_name in required_children,
                    ref_chain=ref_chain,
                )
                if child_fields:
                    child_fields[0].position = index
                fields.extend(child_fields)
    elif type_name == "array" and "items" in schema:
        fields.extend(
            _flatten_openapi_schema(
                schema["items"],
                f"{path}[]",
                resolver=resolver,
                required=True,
                ref_chain=ref_chain,
            )
        )
    return fields


def _merge_openapi_all_of(parts: Any, resolver: _OpenApiResolver) -> dict[str, Any]:
    merged: dict[str, Any] = {"type": "object", "properties": {}, "required": []}
    if not isinstance(parts, list):
        return merged
    for part in parts:
        if not isinstance(part, dict):
            continue
        if "$ref" in part:
            part = resolver.resolve_ref(part["$ref"])
        if "properties" in part and isinstance(part["properties"], dict):
            merged["properties"].update(part["properties"])
        if "required" in part and isinstance(part["required"], list):
            merged["required"].extend(part["required"])
    merged["required"] = sorted(set(merged["required"]))
    return merged


def _openapi_type_name(schema: dict[str, Any]) -> str:
    if "oneOf" in schema or "anyOf" in schema:
        return "mixed"
    schema_type = schema.get("type")
    if isinstance(schema_type, list):
        non_null = {item for item in schema_type if item != "null"}
        if len(non_null) == 1:
            schema_type = next(iter(non_null))
        else:
            return "mixed"
    mapping = {
        "boolean": "boolean",
        "integer": "integer",
        "number": "number",
        "string": "string",
        "array": "array",
        "object": "object",
    }
    if schema_type in mapping:
        return mapping[schema_type]
    if "properties" in schema:
        return "object"
    if "items" in schema:
        return "array"
    return "string"


def _register_avro_types(schema: Any, named_types: dict[str, Any]) -> None:
    if isinstance(schema, dict):
        schema_type = schema.get("type")
        name = schema.get("name")
        if isinstance(name, str) and schema_type in {"record", "enum", "fixed"}:
            named_types[name] = schema
        if schema_type == "record":
            for field in schema.get("fields", []):
                _register_avro_types(field.get("type"), named_types)
        elif schema_type == "array":
            _register_avro_types(schema.get("items"), named_types)
        elif schema_type == "map":
            _register_avro_types(schema.get("values"), named_types)
    elif isinstance(schema, list):
        for part in schema:
            _register_avro_types(part, named_types)


def _flatten_avro_schema(
    schema: Any, path: str, named_types: dict[str, Any], required: bool
) -> list[FieldSchema]:
    resolved = _resolve_avro_type(schema, named_types)
    nullable = not required
    if isinstance(resolved, list):
        non_null = [part for part in resolved if part != "null"]
        nullable = True
        if len(non_null) == 1:
            return _flatten_avro_schema(non_null[0], path, named_types=named_types, required=False)
        return [FieldSchema(path=path, type_name="mixed", nullable=True)]

    if isinstance(resolved, str):
        return [FieldSchema(path=path, type_name=_avro_primitive_type(resolved), nullable=nullable)]

    if not isinstance(resolved, dict):
        return [FieldSchema(path=path, type_name="string", nullable=nullable)]

    schema_type = resolved.get("type")
    if schema_type == "record":
        fields: list[FieldSchema] = []
        if path:
            fields.append(FieldSchema(path=path, type_name="object", nullable=nullable))
        for index, avro_field in enumerate(resolved.get("fields", [])):
            child_path = avro_field["name"] if not path else f"{path}.{avro_field['name']}"
            child_fields = _flatten_avro_schema(
                avro_field.get("type"),
                child_path,
                named_types=named_types,
                required=True,
            )
            if child_fields:
                child_fields[0].position = index
            fields.extend(child_fields)
        return fields

    if schema_type == "array":
        fields = [FieldSchema(path=path, type_name="array", nullable=nullable)]
        fields.extend(
            _flatten_avro_schema(
                resolved.get("items"), f"{path}[]", named_types=named_types, required=True
            )
        )
        return fields

    if schema_type == "map":
        fields = [FieldSchema(path=path, type_name="object", nullable=nullable)]
        fields.extend(
            _flatten_avro_schema(
                resolved.get("values"),
                f"{path}.value",
                named_types=named_types,
                required=True,
            )
        )
        return fields

    if schema_type == "enum":
        samples = [str(value) for value in resolved.get("symbols", [])[:3]]
        return [
            FieldSchema(
                path=path,
                type_name="string",
                nullable=nullable,
                format_hint="enum",
                sample_values=samples,
            )
        ]

    return [
        FieldSchema(
            path=path,
            type_name=_avro_primitive_type(str(schema_type)),
            nullable=nullable,
        )
    ]


def _resolve_avro_type(schema: Any, named_types: dict[str, Any]) -> Any:
    if isinstance(schema, str) and schema in named_types:
        return named_types[schema]
    if isinstance(schema, dict):
        schema_type = schema.get("type")
        if isinstance(schema_type, str) and schema_type in named_types:
            return named_types[schema_type]
    return schema


def _avro_primitive_type(schema_type: str) -> str:
    mapping = {
        "null": "null",
        "boolean": "boolean",
        "int": "integer",
        "long": "integer",
        "float": "number",
        "double": "number",
        "bytes": "binary",
        "string": "string",
        "record": "object",
        "array": "array",
        "map": "object",
        "fixed": "binary",
        "enum": "string",
    }
    return mapping.get(schema_type, "string")


@dataclass
class _ProtoField:
    label: str | None
    type_name: str
    name: str


@dataclass
class _ProtoEnum:
    name: str
    values: list[str] = field(default_factory=list)


@dataclass
class _ProtoMessage:
    name: str
    fields: list[_ProtoField] = field(default_factory=list)
    messages: list[_ProtoMessage] = field(default_factory=list)
    enums: list[_ProtoEnum] = field(default_factory=list)


def _parse_proto(text: str) -> list[_ProtoMessage]:
    lines = _strip_proto_comments(text)
    messages, _ = _parse_proto_block(lines, 0)
    return messages


def _strip_proto_comments(text: str) -> list[str]:
    without_block_comments = re.sub(r"/\*.*?\*/", "", text, flags=re.S)
    lines = []
    for raw_line in without_block_comments.splitlines():
        line = re.sub(r"//.*", "", raw_line).strip()
        if line:
            lines.append(line)
    return lines


def _parse_proto_block(lines: list[str], index: int) -> tuple[list[_ProtoMessage], int]:
    messages: list[_ProtoMessage] = []
    while index < len(lines):
        line = lines[index]
        if line.startswith("}"):
            return messages, index + 1
        if line.startswith("message "):
            match = re.match(r"message\s+(\w+)\s*\{", line)
            if match is None:
                index += 1
                continue
            message_name = match.group(1)
            message, index = _parse_proto_message(lines, index + 1, message_name)
            messages.append(message)
            continue
        index += 1
    return messages, index


def _parse_proto_message(lines: list[str], index: int, name: str) -> tuple[_ProtoMessage, int]:
    message = _ProtoMessage(name=name)
    while index < len(lines):
        line = lines[index]
        if line.startswith("}"):
            return message, index + 1
        if line.startswith("message "):
            nested_name = re.match(r"message\s+(\w+)\s*\{", line)
            if nested_name is not None:
                nested_message, index = _parse_proto_message(lines, index + 1, nested_name.group(1))
                message.messages.append(nested_message)
                continue
        if line.startswith("enum "):
            enum_name = re.match(r"enum\s+(\w+)\s*\{", line)
            if enum_name is not None:
                proto_enum, index = _parse_proto_enum(lines, index + 1, enum_name.group(1))
                message.enums.append(proto_enum)
                continue
        map_match = PROTO_MAP_RE.match(line)
        if map_match is not None:
            label, _, value_type, field_name = map_match.groups()
            message.fields.append(
                _ProtoField(
                    label=label or "repeated", type_name=f"map:{value_type}", name=field_name
                )
            )
            index += 1
            continue
        field_match = PROTO_FIELD_RE.match(line)
        if field_match is not None:
            label, field_type, field_name = field_match.groups()
            message.fields.append(_ProtoField(label=label, type_name=field_type, name=field_name))
        index += 1
    return message, index


def _parse_proto_enum(lines: list[str], index: int, name: str) -> tuple[_ProtoEnum, int]:
    proto_enum = _ProtoEnum(name=name)
    while index < len(lines):
        line = lines[index]
        if line.startswith("}"):
            return proto_enum, index + 1
        enum_value = re.match(r"(\w+)\s*=\s*\d+", line)
        if enum_value is not None:
            proto_enum.values.append(enum_value.group(1))
        index += 1
    return proto_enum, index


def _collect_proto_enums(messages: list[_ProtoMessage]) -> dict[str, list[str]]:
    enum_lookup: dict[str, list[str]] = {}
    for message in messages:
        for proto_enum in message.enums:
            enum_lookup[proto_enum.name] = proto_enum.values
        enum_lookup.update(_collect_proto_enums(message.messages))
    return enum_lookup


def _flatten_proto_message(
    message: _ProtoMessage,
    path: str,
    message_lookup: dict[str, _ProtoMessage],
    enum_lookup: dict[str, list[str]],
    stack: tuple[str, ...],
) -> list[FieldSchema]:
    fields = [FieldSchema(path=path, type_name="object", nullable=False)]
    nested_lookup = {
        **message_lookup,
        **{nested.name: nested for nested in message.messages},
    }
    for index, proto_field in enumerate(message.fields):
        field_path = f"{path}.{proto_field.name}"
        fields.extend(
            _flatten_proto_field(
                proto_field,
                field_path=field_path,
                position=index,
                message_lookup=nested_lookup,
                enum_lookup=enum_lookup,
                stack=(*stack, path),
            )
        )
    for nested in message.messages:
        message_lookup[nested.name] = nested
    return fields


def _flatten_proto_field(
    proto_field: _ProtoField,
    field_path: str,
    position: int,
    message_lookup: dict[str, _ProtoMessage],
    enum_lookup: dict[str, list[str]],
    stack: tuple[str, ...],
) -> list[FieldSchema]:
    nullable = proto_field.label != "required"
    field_type = proto_field.type_name

    if field_type.startswith("map:"):
        value_type = field_type.split(":", maxsplit=1)[1]
        fields = [
            FieldSchema(path=field_path, type_name="object", nullable=True, position=position)
        ]
        fields.extend(
            _flatten_proto_scalar_or_message(
                value_type,
                f"{field_path}.value",
                nullable=False,
                position=None,
                message_lookup=message_lookup,
                enum_lookup=enum_lookup,
                stack=stack,
            )
        )
        return fields

    if proto_field.label == "repeated":
        fields = [FieldSchema(path=field_path, type_name="array", nullable=True, position=position)]
        fields.extend(
            _flatten_proto_scalar_or_message(
                field_type,
                f"{field_path}[]",
                nullable=False,
                position=None,
                message_lookup=message_lookup,
                enum_lookup=enum_lookup,
                stack=stack,
            )
        )
        return fields

    return _flatten_proto_scalar_or_message(
        field_type,
        field_path,
        nullable=nullable,
        position=position,
        message_lookup=message_lookup,
        enum_lookup=enum_lookup,
        stack=stack,
    )


def _flatten_proto_scalar_or_message(
    field_type: str,
    field_path: str,
    nullable: bool,
    position: int | None,
    message_lookup: dict[str, _ProtoMessage],
    enum_lookup: dict[str, list[str]],
    stack: tuple[str, ...],
) -> list[FieldSchema]:
    normalized = _proto_scalar_type(field_type)
    if normalized is not None:
        return [
            FieldSchema(
                path=field_path,
                type_name=normalized,
                nullable=nullable,
                position=position,
            )
        ]
    if field_type in enum_lookup:
        return [
            FieldSchema(
                path=field_path,
                type_name="string",
                nullable=nullable,
                position=position,
                format_hint="enum",
                sample_values=enum_lookup[field_type][:3],
            )
        ]
    if field_type in message_lookup:
        message = message_lookup[field_type]
        if any(ancestor.split(".")[-1] == message.name for ancestor in stack):
            return [
                FieldSchema(
                    path=field_path,
                    type_name="object",
                    nullable=nullable,
                    position=position,
                )
            ]
        fields = [
            FieldSchema(path=field_path, type_name="object", nullable=nullable, position=position)
        ]
        for index, proto_field in enumerate(message.fields):
            child_path = f"{field_path}.{proto_field.name}"
            fields.extend(
                _flatten_proto_field(
                    proto_field,
                    field_path=child_path,
                    position=index,
                    message_lookup=message_lookup,
                    enum_lookup=enum_lookup,
                    stack=(*stack, field_path),
                )
            )
        return fields
    return [FieldSchema(path=field_path, type_name="string", nullable=nullable, position=position)]


def _proto_scalar_type(field_type: str) -> str | None:
    mapping = {
        "bool": "boolean",
        "bytes": "binary",
        "double": "number",
        "fixed32": "integer",
        "fixed64": "integer",
        "float": "number",
        "int32": "integer",
        "int64": "integer",
        "sfixed32": "integer",
        "sfixed64": "integer",
        "sint32": "integer",
        "sint64": "integer",
        "string": "string",
        "uint32": "integer",
        "uint64": "integer",
    }
    return mapping.get(field_type)


def _safe_segment(value: str) -> str:
    sanitized = value.strip("/")
    sanitized = sanitized.replace("/", ".")
    sanitized = re.sub(r"[^A-Za-z0-9_.{}-]+", "_", sanitized)
    return sanitized or "root"


class _OpenApiResolver:
    def __init__(self, payload: dict[str, Any]) -> None:
        self.payload = payload

    def resolve_ref(self, ref: str) -> dict[str, Any]:
        if not ref.startswith("#/"):
            raise ValueError(f"Only local OpenAPI refs are supported: {ref}")
        current: Any = self.payload
        for segment in ref[2:].split("/"):
            current = current[segment]
        if not isinstance(current, dict):
            raise ValueError(f"OpenAPI ref {ref} did not resolve to an object")
        return current


__author__ = "gadwant"



from __future__ import annotations

import json
from pathlib import Path

import pyarrow as pa
import pyarrow.parquet as pq
from typer.testing import CliRunner

from schemaglow.cli import app

runner = CliRunner()


def _write(path: Path, content: str) -> Path:
    path.write_text(content, encoding="utf-8")
    return path


def test_diff_command_outputs_json_and_writes_html_report_for_parquet(tmp_path: Path) -> None:
    old_file = tmp_path / "old.parquet"
    new_file = tmp_path / "new.parquet"
    report_path = tmp_path / "report.html"

    pq.write_table(pa.table({"id": [1, 2]}), old_file)
    pq.write_table(pa.table({"id": [1, 2], "campaign_id": ["a", None]}), new_file)

    result = runner.invoke(
        app,
        [
            "diff",
            str(old_file),
            str(new_file),
            "--format",
            "json",
            "--report",
            "html",
            "--report-path",
            str(report_path),
        ],
    )

    assert result.exit_code == 0
    payload = json.loads(result.stdout.split("report written to", maxsplit=1)[0].strip())
    assert payload["overall"] == "SAFE"
    assert any(event["code"] == "field_added_optional" for event in payload["events"])
    assert report_path.exists()
    assert "<html" in report_path.read_text(encoding="utf-8").lower()


def test_inspect_snapshot_and_compare_roundtrip(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "id,total\n1,10\n2,20\n")
    new_file = _write(tmp_path / "new.csv", "id,total,status\n1,10,open\n2,20,\n")
    old_snapshot = tmp_path / "old.schema.json"
    new_snapshot = tmp_path / "new.schema.json"

    inspect_result = runner.invoke(app, ["inspect", str(old_file)])
    assert inspect_result.exit_code == 0
    assert "SchemaGlow Inspect" in inspect_result.stdout
    assert "- id: integer (required)" in inspect_result.stdout

    snapshot_old_result = runner.invoke(app, ["snapshot", str(old_file), "-o", str(old_snapshot)])
    snapshot_new_result = runner.invoke(app, ["snapshot", str(new_file), "-o", str(new_snapshot)])
    assert snapshot_old_result.exit_code == 0
    assert snapshot_new_result.exit_code == 0
    assert old_snapshot.exists()
    assert new_snapshot.exists()

    compare_result = runner.invoke(
        app, ["compare", str(old_snapshot), str(new_snapshot), "--format", "json"]
    )
    assert compare_result.exit_code == 0
    payload = json.loads(compare_result.stdout)
    assert payload["overall"] == "SAFE"
    assert any(event["path"] == "status" for event in payload["events"])


def test_diff_markdown_stdout_and_inspect_json_output(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.json", '{"order":{"id":1,"total":10}}')
    new_file = _write(tmp_path / "new.json", '{"order":{"id":1,"total":10,"status":"open"}}')

    diff_result = runner.invoke(
        app,
        [
            "diff",
            str(old_file),
            str(new_file),
            "--report",
            "markdown",
        ],
    )
    inspect_result = runner.invoke(app, ["inspect", str(new_file), "--format", "json"])

    assert diff_result.exit_code == 0
    assert "SchemaGlow Report" in diff_result.stdout
    assert "# SchemaGlow Report" in diff_result.stdout
    assert inspect_result.exit_code == 0
    payload = json.loads(inspect_result.stdout)
    assert payload["source_format"] == "json"
    assert any(field["path"] == "order.status" for field in payload["fields"])


def test_compare_text_output_and_invalid_report_value(tmp_path: Path) -> None:
    old_file = _write(tmp_path / "old.csv", "id,total\n1,10\n")
    new_file = _write(tmp_path / "new.csv", "id,total\n1,10.5\n")
    old_snapshot = tmp_path / "old.schema.json"
    new_snapshot = tmp_path / "new.schema.json"

    runner.invoke(app, ["snapshot", str(old_file), "-o", str(old_snapshot)])
    runner.invoke(app, ["snapshot", str(new_file), "-o", str(new_snapshot)])

    compare_result = runner.invoke(app, ["compare", str(old_snapshot), str(new_snapshot)])
    invalid_report = runner.invoke(app, ["diff", str(old_file), str(new_file), "--report", "xml"])

    assert compare_result.exit_code == 0
    assert "SchemaGlow Report" in compare_result.stdout
    assert invalid_report.exit_code != 0


def test_scan_command_and_baseline_commands(tmp_path: Path) -> None:
    old_dir = tmp_path / "old"
    new_dir = tmp_path / "new"
    baseline_dir = tmp_path / ".schemaglow-baseline"
    old_dir.mkdir()
    new_dir.mkdir()

    _write(old_dir / "orders.csv", "id,total\n1,10\n")
    _write(new_dir / "orders.csv", "id,total,status\n1,10,\n")
    _write(
        new_dir / "openapi.yaml",
        "openapi: 3.1.0\ninfo:\n  title: Demo\n  version: '1.0'\npaths: {}\n",
    )

    scan_result = runner.invoke(app, ["scan", str(old_dir), str(new_dir), "--format", "json"])
    capture_result = runner.invoke(
        app, ["baseline", "capture", str(old_dir), "-o", str(baseline_dir)]
    )
    check_result = runner.invoke(
        app,
        ["baseline", "check", str(baseline_dir), str(new_dir), "--format", "json"],
    )

    assert scan_result.exit_code == 0
    scan_payload = json.loads(scan_result.stdout)
    assert scan_payload["overall"] == "WARNING"
    assert any(entry["status"] == "added_file" for entry in scan_payload["entries"])

    assert capture_result.exit_code == 0
    assert "SchemaGlow Baseline" in capture_result.stdout
    assert (baseline_dir / "schemaglow-baseline.json").exists()

    assert check_result.exit_code == 0
    check_payload = json.loads(check_result.stdout)
    assert any(entry["status"] == "unexpected_candidate" for entry in check_payload["entries"])
